import numpy as np
import torch
import torchvision
from torch.fft import fft2, fftshift, ifft2
from torchvision.io import read_image
from torchvision.transforms.functional import resize, rgb_to_grayscale

from .utils import cart2pol_torch, denoise_torch, morph_torch


class PAGE_GPU:
    def __init__(self, direction_bins, device, h=None, w=None):

        """initialize the PAGE GPU version class

        Args:
            direction_bins (int): number of different diretions of edge to be extracted
            device(torch.device)
            h (int, optional): height of the image to be processed. Defaults to None.
            w (int, optional): width of the image to be processed. Defaults to None.
        """
        self.h = h
        self.w = w
        self.direction_bins = direction_bins
        self.device = device

    def load_img(self, img_file=None, img_array=None):
        """load the image from an ndarray or from an image file

        Args:
            img_file (str, optional): path to the image. Defaults to None.
            img_array (torch.Tensor, optional): image in the form of torch.Tensor. Defaults to None.
        """
        if img_array is not None:
            # directly load the image from the array instead of the file
            if img_array.get_device() == self.device:
                self.img = img_array
            else:    
                self.img = img_array.to(self.device)
            # convert to grayscale if it is RGB
            if self.img.dim() == 3 and self.img.shape[0] != 1:
                self.img = rgb_to_grayscale(self.img)
            # read the image size or resize to the indicated size (height x width)
            if not self.h and not self.w:
                self.img = torch.squeeze(self.img)
                self.h = self.img.shape[0]
                self.w = self.img.shape[1]
            else:
                self.img = torch.squeeze(resize(self.img, [self.h, self.w]))

        else:
            # load the image from the image file
            # torchvision read_image only supports 'jpg' and 'png'
            if img_file.split(".")[-1] in ["jpg", "png", "jpeg"]:
                self.img = torchvision.io.read_image(img_file).to(self.device)
                # convert to grayscale if it is RGB
                if self.img.dim() == 3 and self.img.shape[0] != 1:
                    self.img = rgb_to_grayscale(self.img)
                # read the image size or resize to the indicated size (height x width)
                if not self.h and not self.w:
                    self.img = torch.squeeze(self.img)
                    self.h = self.img.shape[0]
                    self.w = self.img.shape[1]
                else:
                    self.img = torch.squeeze(resize(self.img, [self.h, self.w]))

            else:
                # use opencv to load other format of image
                self.img = cv2.imread(img_file)
                if self.img.ndim == 3:
                    self.img = cv2.cvtColor(self.img, cv2.COLOR_BGR2GRAY)
                if not self.h and not self.w:
                    self.h = self.img.shape[0]
                    self.w = self.img.shape[1]
                else:
                    self.img = cv2.imresize(self.img, [self.h, self.w])

                self.img = torch.from_numpy(self.img).to(self.device)

    def init_kernel(self, mu_1, mu_2, sigma_1, sigma_2, S1, S2):
        """initialize the phase kernel of PST

        Args:
            mu_1 (float): Center frequency of a normal distributed passband filter ϕ1
            mu_2 (float):  Center frequency of log-normal  distributed passband filter ϕ2
            sigma_1 (float): Standard deviation of normal distributed passband filter ϕ1
            sigma_2 (float): Standard deviation of log-normal distributed passband filter ϕ2
            S1 (float): Phase strength of ϕ1
            S2 (float): Phase strength of ϕ2
        """

        # set the frequency grid
        u = torch.linspace(-0.5, 0.5, self.h, device=self.device).float()
        v = torch.linspace(-0.5, 0.5, self.w, device=self.device).float()
        [U, V] = torch.meshgrid(u, v, indexing="ij")
        [self.THETA, self.RHO] = cart2pol_torch(U, V)

        min_direction = np.pi / 180
        direction_span = np.pi / self.direction_bins
        directions = torch.arange(min_direction, np.pi, direction_span)

        # create PAGE kernels in parallel by broadcasting
        tetavs = torch.unsqueeze(directions, dim=0).to(self.device)
        Us = torch.unsqueeze(U, dim=-1)
        Vs = torch.unsqueeze(V, dim=-1)
        Uprimes = Us * torch.cos(tetavs) + Vs * torch.sin(tetavs)
        Vprimes = -Us * torch.sin(tetavs) + Vs * torch.cos(tetavs)

        Phi_1s = torch.exp(-0.5 * ((torch.abs(Uprimes) - mu_1) / sigma_1) ** 2) / (
            1 * np.sqrt(2 * np.pi) * sigma_1
        )
        Phi_1s = (
            Phi_1s / torch.max(Phi_1s.view(-1, self.direction_bins), dim=0)[0]
        ) * S1

        Phi_2s = torch.exp(
            -0.5 * ((torch.log(torch.abs(Vprimes)) - mu_2) / sigma_2) ** 2
        ) / (abs(Vprimes) * np.sqrt(2 * np.pi) * sigma_2)
        Phi_2s = (
            Phi_2s / torch.max(Phi_2s.view(-1, self.direction_bins), dim=0)[0]
        ) * S2
        self.page_kernel = Phi_1s * Phi_2s

    def apply_kernel(self, sigma_LPF, thresh_min, thresh_max, morph_flag):

        """apply the phase kernel onto the image
        Args:
            sigma_LPF (float): std of the low pass filter
            thresh_min (float): minimum thershold, we keep features < thresh_min
            thresh_max (float): maximum thershold, we keep features > thresh_max
            morph_flag (boolean): whether apply morphological operation
        """
        # denoise on the loaded image
        self.img_denoised = denoise_torch(
            img=self.img, rho=self.RHO, sigma_LPF=sigma_LPF
        )
        # apply the page kernel
        self.img_page = ifft2(
            fft2(self.img_denoised).unsqueeze(-1)
            * fftshift(torch.exp(-1j * self.page_kernel), dim=(0, 1)),
            dim=(0, 1),
        )
        self.page_feature = torch.angle(self.img_page)
        # apply morphological operation if applicable
        if morph_flag == 0:
            self.page_output = self.page_feature
        else:
            kernel = torch.tensor(
                [[0.0, 1.0, 0.0], [1.0, 1.0, 1.0], [0.0, 1.0, 0.0]]
            ).to(self.device)

            self.page_output = morph_torch(
                img=self.img,
                feature=self.page_feature,
                thresh_min=thresh_min,
                thresh_max=thresh_max,
                kernel=kernel,
                device=self.device,
            )

    def create_page_edge(self):
        """create results which color-coded directional edges"""
        # Create a weighted color image of PAGE output to visualize directionality of edges
        weight_step = 255 * 3 / self.direction_bins
        color_weight = torch.arange(0, 255, weight_step).to(self.device)
        self.page_edge = torch.zeros([self.h, self.w, 3]).to(self.device)
        # step_edge = int(round(self.direction_bins/3))
        step_edge = self.direction_bins // 3
        for i in range(step_edge):
            self.page_edge[:, :, 0] = (
                color_weight[i] * self.page_output[:, :, i] + self.page_edge[:, :, 0]
            )
            self.page_edge[:, :, 1] = (
                color_weight[i] * self.page_output[:, :, i + step_edge]
                + self.page_edge[:, :, 1]
            )
            self.page_edge[:, :, 2] = (
                color_weight[i] * self.page_output[:, :, i + (2 * step_edge)]
                + self.page_edge[:, :, 2]
            )

        self.page_edge = (self.page_edge - torch.min(self.page_edge)) / (
            torch.max(self.page_edge) - torch.min(self.page_edge)
        )

    def run(
        self,
        img_file,
        mu_1,
        mu_2,
        sigma_1,
        sigma_2,
        S1,
        S2,
        sigma_LPF,
        thresh_min,
        thresh_max,
        morph_flag,
    ):
        """wrap all steps of PAGE into a single run method

        Args:
            img_file (str): path to the image.
            mu_1 (float): Center frequency of a normal distributed passband filter ϕ1
            mu_2 (float):  Center frequency of log-normal  distributed passband filter ϕ2
            sigma_1 (float): Standard deviation of normal distributed passband filter ϕ1
            sigma_2 (float): Standard deviation of log-normal distributed passband filter ϕ2
            S1 (float): Phase strength of ϕ1
            S2 (float): Phase strength of ϕ2
            sigma_LPF (float): std of the low pass filter
            thresh_min (float): minimum thershold, we keep features < thresh_min
            thresh_max (float): maximum thershold, we keep features > thresh_max
            morph_flag (boolean): whether apply morphological operation

        Returns:
            torch.Tensor: color-coded directional edge
        """
        self.load_img(img_file=img_file)
        self.init_kernel(mu_1, mu_2, sigma_1, sigma_2, S1, S2)
        self.apply_kernel(sigma_LPF, thresh_min, thresh_max, morph_flag)
        self.create_page_edge()

        return self.page_edge



import cv2
import kornia
import mahotas as mh
import numpy as np
import torch
import torch.fft


def cart2pol(x, y):
    """convert cartesian coordiates to polar coordinates

    Args:
        x (np.ndarray): cartesian coordinates in x direction
        y (np.ndarray): cartesian coordinates in y direction

    Returns:
        tuple: polar coordinates theta and rho
    """
    theta = np.arctan2(y, x)
    rho = np.hypot(x, y)
    return (theta, rho)


def cart2pol_torch(x, y):
    """convert cartesian coordiates to polar coordinates with PyTorch

    Args:
        x (torch.Tensor): cartesian coordinates in x direction
        y (torch.Tensor): cartesian coordinates in x direction

    Returns:
        tuple: polar coordinates theta and rho
    """
    theta = torch.atan2(y, x)
    rho = torch.hypot(x, y)
    return (theta, rho)


def denoise(img, rho, sigma_LPF):
    """apply a low pass filter to denoise the image

    Args:
        img (np.ndarray): original image
        rho (np.ndarray): polar coordinates
        sigma_LPF (float): variance of the low pass filter

    Returns:
        np.ndarray: denoised image
    """
    img_orig_f = np.fft.fft2(img)
    expo = np.fft.fftshift(
        np.exp(
            -0.5 * np.power((np.divide(rho, np.sqrt((sigma_LPF**2) / np.log(2)))), 2)
        )
    )
    img_filtered = np.real(np.fft.ifft2((np.multiply(img_orig_f, expo))))

    return img_filtered


def denoise_torch(img, rho, sigma_LPF):
    """apply a low pass filter to denoise the image with PyTorch

    Args:
        img (torch.Tensor): original image
        rho (torch.Tensor): polar coordinates
        sigma_LPF (float): std of the low pass filter

    Returns:
        torch.Tensor: denoised image
    """
    img_orig_f = torch.fft.fft2(img)
    expo = torch.fft.fftshift(
        torch.exp(
            -0.5
            * torch.pow((torch.divide(rho, np.sqrt((sigma_LPF**2) / np.log(2)))), 2)
        )
    )
    img_filtered = torch.real(torch.fft.ifft2((torch.mul(img_orig_f, expo))))

    return img_filtered


def morph(img, feature, thresh_min, thresh_max):
    """apply morphological operation to transform analog features to digial features

    Args:
        img (np.ndarray): original image
        feature (np.ndarray): analog feature
        thresh_min (float): minimum thershold, we keep features < thresh_min
        thresh_max (float): maximum thershold, we keep features > thresh_max

    Returns:
        np.ndarray: digital features (binary edge)
    """

    digital_feature = np.zeros(feature.shape)

    digital_feature[feature > thresh_max] = 1
    digital_feature[feature < thresh_min] = 1
    digital_feature[img < (np.amax(img) / 20)] = 0

    digital_feature = mh.thin(digital_feature, 1)
    digital_feature = mh.bwperim(digital_feature, 4)
    digital_feature = mh.thin(digital_feature, 1)
    digital_feature = mh.erode(digital_feature, np.ones((1, 1)))

    return digital_feature.astype(np.float32)


def morph_torch(img, feature, thresh_min, thresh_max, kernel, device):
    """apply morphological operation to transform analog features to digial features in PyTorch

    Args:
        img (torch.Tensor): original image
        feature (torch.Tensor): analog feature
        thresh_min (torch.Tensor): minimum thershold, we keep features < thresh_min
        thresh_max (torch.Tensor): maximum thershold, we keep features > thresh_max
        device (torch.device)

    Returns:
        torch.Tensor: digital features (binary edge)
    """
    digital_feature = torch.zeros(feature.shape).to(device)
    digital_feature[feature > thresh_max] = 1
    digital_feature[feature < thresh_min] = 1
    digital_feature[img < (torch.max(img) / 20)] = 0

    # for PST
    if len(feature.shape) == 2:
        digital_feature = kornia.utils.image._to_bchw(digital_feature)
        digital_feature = kornia.morphology.closing(digital_feature, kernel)

    # for PAGE
    elif len(feature.shape) == 3:
        digital_feature = torch.permute(digital_feature, (2, 0, 1))
        digital_feature = kornia.utils.image._to_bchw(
            torch.unsqueeze(digital_feature, 1)
        )
        digital_feature = torch.squeeze(
            kornia.morphology.closing(digital_feature, kernel)
        )
        digital_feature = torch.permute(digital_feature, (1, 2, 0))

    return torch.squeeze(digital_feature)



import cv2
import numpy as np
from numpy.fft import fft2, fftshift, ifft2

from .utils import cart2pol


class VEVID:
    def __init__(self, h=None, w=None):
        """initialize the VEVID CPU version class

        Args:
            h (int, optional): height of the image to be processed. Defaults to None.
            w (int, optional): width of the image to be processed. Defaults to None.
        """
        self.h = h
        self.w = w

    def load_img(self, img_file=None, img_array=None):
        """load the image from an ndarray or from an image file

        Args:
            img_file (str, optional): path to the image. Defaults to None.
            img_array (np.ndarray, optional): image in the form of np.ndarray. Defaults to None.
        """
        if img_array is not None:
            # directly load the image from numpy array
            self.img_bgr = img_array
        else:
            # load the image from the image file
            self.img_bgr = cv2.imread(img_file)
            if not self.h and not self.w:
                self.h = self.img_bgr.shape[0]
                self.w = self.img_bgr.shape[1]
            else:
                self.img_bgr = cv2.resize(self.img_bgr, [self.w, self.h])

        self.img_hsv = cv2.cvtColor(self.img_bgr, cv2.COLOR_BGR2HSV) / 255.0

    def init_kernel(self, S, T):
        """initialize the phase kernel of VEViD

        Args:
            S (float): phase strength
            T (float): variance of the spectral phase function
        """
        # create the frequency grid
        u = np.linspace(-0.5, 0.5, self.h)
        v = np.linspace(-0.5, 0.5, self.w)
        [U, V] = np.meshgrid(u, v, indexing="ij")
        # construct the kernel
        [self.THETA, self.RHO] = cart2pol(U, V)
        self.vevid_kernel = np.exp(-self.RHO**2 / T)
        self.vevid_kernel = (self.vevid_kernel / np.max(abs(self.vevid_kernel))) * S

    def apply_kernel(self, b, G, color=False, lite=False):
        """apply the phase kernel onto the image

        Args:
            b (float): regularization term
            G (float): phase activation gain
            color (bool, optional): whether to run color enhancement. Defaults to False.
            lite (bool, optional): whether to run VEViD lite. Defaults to False.
        """
        if color:
            channel_idx = 1
        else:
            channel_idx = 2
        vevid_input = self.img_hsv[:, :, channel_idx]
        if lite:
            vevid_phase = np.arctan2(-G * (vevid_input + b), vevid_input)
        else:
            vevid_input_f = fft2(vevid_input + b)
            img_vevid = ifft2(vevid_input_f * fftshift(np.exp(-1j * self.vevid_kernel)))
            vevid_phase = np.arctan2(G * np.imag(img_vevid), vevid_input)
        vevid_phase_norm = (vevid_phase - vevid_phase.min()) / (
            vevid_phase.max() - vevid_phase.min()
        )
        self.img_hsv[:, :, channel_idx] = vevid_phase_norm
        self.img_hsv = cv2.normalize(
            self.img_hsv, None, 0, 255, cv2.NORM_MINMAX
        ).astype(np.uint8)
        self.vevid_output = cv2.cvtColor(self.img_hsv, cv2.COLOR_HSV2RGB)

    def run(self, img_file, S, T, b, G, color=False):
        """run the full VEViD algorithm

        Args:
            img_file (str): path to the image
            S (float): phase strength
            T (float): variance of the spectral phase function
            b (float): regularization term
            G (float): phase activation gain
            color (bool, optional): whether to run color enhancement. Defaults to False.

        Returns:
            np.ndarray: enhanced image
        """
        self.load_img(img_file=img_file)
        self.init_kernel(S, T)
        self.apply_kernel(b, G, color, lite=False)

        return self.vevid_output

    def run_lite(self, img_file, b, G, color=False):
        """run the VEViD lite algorithm

        Args:
            img_file (str): path to the image
            b (float): regularization term
            G (float): phase activation gain
            color (bool, optional): whether to run color enhancement. Defaults to False.

        Returns:
            np.ndarray: enhanced image
        """
        self.load_img(img_file=img_file)
        self.apply_kernel(b, G, color, lite=True)

        return self.vevid_output



import cv2
import numpy as np
import torch
from kornia.color import hsv_to_rgb, rgb_to_hsv
from torch.fft import fft2, fftshift, ifft2
from torchvision.io import read_image
from torchvision.transforms.functional import resize, rgb_to_grayscale

from .utils import cart2pol_torch


class VEVID_GPU:
    def __init__(self, device, h=None, w=None):
        """initialize the VEVID GPU version class

        Args:
            device (torch.device)
            h (int, optional): height of the image to be processed. Defaults to None.
            w (int, optional): width of the image to be processed. Defaults to None.
        """
        self.h = h
        self.w = w
        self.device = device

    def load_img(self, img_file=None, img_array=None):
        """load the image from an ndarray or from an image file

        Args:
            img_file (str, optional): path to the image. Defaults to None.
            img_array (torch.Tensor, optional): image in the form of torch.Tensor. Defaults to None.
        """
        if img_array is not None:
            # directly load the image from the array instead of the file
            if img_array.get_device() == self.device:
                self.img_rgb = img_array
            else:    
                self.img_rgb = img_array.to(self.device)
            if not self.h and not self.w:
                self.h = self.img_rgb.shape[-2]
                self.w = self.img_rgb.shape[-1]
            # convert from RGB to HSV
            self.img_hsv = rgb_to_hsv(self.img_rgb)

        else:
            # load the image from the image file
            # torchvision read_image currently only supports 'jpg' and 'png'
            # use opencv to read other image formats
            if img_file.split(".")[-1] in ["jpg", "png", "jpeg"]:
                self.img_rgb = read_image(img_file).to(self.device)
            else:
                self.img_bgr = cv2.imread(img_file)
                self.img_rgb = cv2.cvtColor(self.img_bgr, cv2.COLOR_BGR2RGB)
                self.img_rgb = torch.from_numpy(
                    np.transpose(self.img_rgb, (2, 0, 1))
                ).to(self.device)
            if not self.h and not self.w:
                self.h = self.img_rgb.shape[-2]
                self.w = self.img_rgb.shape[-1]
            else:
                self.img_rgb = resize(self.img_rgb, [self.h, self.w])
            # convert from RGB to HSV
            # rgb_to_hsv in kornia requires the input RGB image to be in the range of 0-1
            self.img_hsv = rgb_to_hsv((self.img_rgb.float()) / 255.0)

    def init_kernel(self, S, T):
        """initialize the phase kernel of VEViD

        Args:
            S (float): phase strength
            T (float): variance of the spectral phase function
        """
        # create the frequency grid
        u = torch.linspace(-0.5, 0.5, self.h, device=self.device).float()
        v = torch.linspace(-0.5, 0.5, self.w, device=self.device).float()
        [U, V] = torch.meshgrid(u, v, indexing="ij")
        # construct the kernel
        [self.THETA, self.RHO] = cart2pol_torch(U, V)
        self.vevid_kernel = torch.exp(-self.RHO**2 / T)
        self.vevid_kernel = (self.vevid_kernel / torch.max(abs(self.vevid_kernel))) * S

    def apply_kernel(self, b, G, color=False, lite=False):
        """apply the phase kernel onto the image

        Args:
            b (float): regularization term
            G (float): phase activation gain
            color (bool, optional): whether to run color enhancement. Defaults to False.
            lite (bool, optional): whether to run VEViD lite. Defaults to False.
        """
        if color:
            channel_idx = 1
        else:
            channel_idx = 2
        vevid_input = self.img_hsv[channel_idx, :, :]
        if lite:
            vevid_phase = torch.atan2(-G * (vevid_input + b), vevid_input)
        else:
            vevid_input_f = fft2(vevid_input + b)
            img_vevid = ifft2(
                vevid_input_f * fftshift(torch.exp(-1j * self.vevid_kernel))
            )
            vevid_phase = torch.atan2(G * torch.imag(img_vevid), vevid_input)
        vevid_phase_norm = (vevid_phase - vevid_phase.min()) / (
            vevid_phase.max() - vevid_phase.min()
        )
        self.img_hsv[channel_idx, :, :] = vevid_phase_norm
        self.vevid_output = hsv_to_rgb(self.img_hsv)

    def run(self, img_file, S, T, b, G, color=False):
        """run the full VEViD algorithm

        Args:
            img_file (str): path to the image
            S (float): phase strength
            T (float): variance of the spectral phase function
            b (float): regularization term
            G (float): phase activation gain
            color (bool, optional): whether to run color enhancement. Defaults to False.

        Returns:
            torch.Tensor: enhanced image
        """
        self.load_img(img_file=img_file)
        self.init_kernel(S, T)
        self.apply_kernel(b, G, color, lite=False)

        return self.vevid_output

    def run_lite(self, img_file, b, G, color=False):
        """run the VEViD lite algorithm

        Args:
            img_file (str): path to the image
            b (float): regularization term
            G (float): phase activation gain
            color (bool, optional): whether to run color enhancement. Defaults to False.

        Returns:
            torch.Tensor: enhanced image
        """
        self.load_img(img_file=img_file)
        self.apply_kernel(b, G, color, lite=True)

        return self.vevid_output



import numpy as np
import torch
import torchvision
from torch.fft import fft2, fftshift, ifft2
from torchvision.io import read_image
from torchvision.transforms.functional import resize, rgb_to_grayscale

from .utils import cart2pol_torch, denoise_torch, morph_torch


class PST_GPU:
    def __init__(self, device, h=None, w=None):
        """initialize the PST GPU version class

        Args:
            device (torch.device)
            h (int, optional): height of the image to be processed. Defaults to None.
            w (int, optional): width of the image to be processed. Defaults to None.
        """
        self.h = h
        self.w = w
        self.device = device

    def load_img(self, img_file=None, img_array=None):
        """load the image from an ndarray or from an image file

        Args:
            img_file (str, optional): path to the image. Defaults to None.
            img_array (torch.Tensor, optional): image in the form of torch.Tensor. Defaults to None.
        """
        if img_array is not None:
            # directly load the image from the array instead of the file
            if img_array.get_device() == self.device:
                self.img = img_array
            else:    
                self.img = img_array.to(self.device)            
            # convert to grayscale if it is RGB
            if self.img.dim() == 3 and self.img.shape[0] != 1:
                self.img = rgb_to_grayscale(self.img)
            # read the image size or resize to the indicated size (height x width)
            if not self.h and not self.w:
                self.img = torch.squeeze(self.img)
                self.h = self.img.shape[0]
                self.w = self.img.shape[1]
            else:
                self.img = torch.squeeze(resize(self.img, [self.h, self.w]))
        else:
            # load the image from the image file
            # torchvision read_image only supports 'jpg' and 'png'
            if img_file.split(".")[-1] in ["jpg", "png", "jpeg"]:
                self.img = torchvision.io.read_image(img_file).to(self.device)
                # convert to grayscale if it is RGB
                if self.img.dim() == 3 and self.img.shape[0] != 1:
                    self.img = rgb_to_grayscale(self.img)
                # read the image size or resize to the indicated size (height x width)
                if not self.h and not self.w:
                    self.img = torch.squeeze(self.img)
                    self.h = self.img.shape[0]
                    self.w = self.img.shape[1]
                else:
                    self.img = torch.squeeze(resize(self.img, [self.h, self.w]))
            else:
                # use opencv to load other format of image
                self.img = cv2.imread(img_file)
                if self.img.ndim == 3:
                    self.img = cv2.cvtColor(self.img, cv2.COLOR_BGR2GRAY)
                if not self.h and not self.w:
                    self.h = self.img.shape[0]
                    self.w = self.img.shape[1]
                else:
                    self.img = cv2.imresize(self.img, [self.h, self.w])

                self.img = torch.from_numpy(self.img).to(self.device)

    def init_kernel(self, S, W):
        """initialize the phase kernel of PST

        Args:
            S (float): phase strength of PST
            W (float): warp of PST
        """
        # set the frequency grid
        u = torch.linspace(-0.5, 0.5, self.h, device=self.device).float()
        v = torch.linspace(-0.5, 0.5, self.w, device=self.device).float()
        [U, V] = torch.meshgrid(u, v, indexing="ij")
        [self.THETA, self.RHO] = cart2pol_torch(U, V)
        # construct the PST Kernel
        self.pst_kernel = W * self.RHO * torch.arctan(W * self.RHO) - 0.5 * torch.log(
            1 + (W * self.RHO) ** 2
        )
        self.pst_kernel = S * self.pst_kernel / torch.max(self.pst_kernel)

    def apply_kernel(self, sigma_LPF, thresh_min, thresh_max, morph_flag):

        """apply the phase kernel onto the image

        Args:
            sigma_LPF (float): std of the low pass filter
            thresh_min (float): minimum thershold, we keep features < thresh_min
            thresh_max (float): maximum thershold, we keep features > thresh_max
            morph_flag (boolean): whether apply morphological operation
        """
        # denoise on the loaded image
        self.img_denoised = denoise_torch(
            img=self.img, rho=self.RHO, sigma_LPF=sigma_LPF
        )
        # apply the pst kernel
        self.img_pst = ifft2(
            fft2(self.img_denoised) * fftshift(torch.exp(-1j * self.pst_kernel))
        )
        self.pst_feature = torch.angle(self.img_pst)
        # apply morphological operation if applicable
        if morph_flag == 0:
            self.pst_output = self.pst_feature
        else:
            kernel = torch.tensor(
                [[0.0, 1.0, 0.0], [1.0, 1.0, 1.0], [0.0, 1.0, 0.0]]
            ).to(self.device)
            self.pst_output = morph_torch(
                img=self.img,
                feature=self.pst_feature,
                thresh_max=thresh_max,
                thresh_min=thresh_min,
                kernel=kernel,
                device=self.device,
            )

    def run(
        self,
        img_file,
        S,
        W,
        sigma_LPF,
        thresh_min,
        thresh_max,
        morph_flag,
    ):
        """wrap all steps of PST into a single run method

        Args:
            img_file (str): _description_
            S (float): _description_
            W (float): _description_
            sigma_LPF (float): _description_
            thresh_min (float): _description_
            thresh_max (float): _description_
            morph_flag (boolean): _description_

        Returns:
            torch.Tensor: PST output
        """
        # wrap load_img, init_kernel, apply_kernel in one run
        self.load_img(img_file=img_file)
        self.init_kernel(S, W)
        self.apply_kernel(sigma_LPF, thresh_min, thresh_max, morph_flag)

        return self.pst_output



import cv2
import numpy as np
from numpy.fft import fft2, fftshift, ifft2

from .utils import cart2pol, denoise, morph


class PST:
    def __init__(self, h=None, w=None):
        """initialize the PST CPU version class

        Args:
            h (int, optional): height of the image to be processed. Defaults to None.
            w (int, optional): width of the image to be processed. Defaults to None.
        """
        self.h = h
        self.w = w

    def load_img(self, img_file=None, img_array=None):
        """load the image from an ndarray or from an image file

        Args:
            img_file (str, optional): path to the image. Defaults to None.
            img_array (np.ndarray, optional): image in the form of np.ndarray. Defaults to None.
        """
        if img_array is not None:
            self.img = img_array
        else:
            self.img = cv2.imread(img_file)
            if not self.h and not self.w:
                self.h = self.img.shape[0]
                self.w = self.img.shape[1]
            else:
                self.img = cv2.imresize(self.img, [self.h, self.w])
        # convert to grayscale if it is RGB
        if self.img.ndim == 3:
            self.img = cv2.cvtColor(self.img, cv2.COLOR_BGR2GRAY)

    def init_kernel(self, S, W):
        """initialize the phase kernel of PST

        Args:
            S (float): phase strength of PST
            W (float): warp strength of PST
        """
        # set the frequency grid
        u = np.linspace(-0.5, 0.5, self.h)
        v = np.linspace(-0.5, 0.5, self.w)
        [U, V] = np.meshgrid(u, v, indexing="ij")
        [self.THETA, self.RHO] = cart2pol(U, V)
        # construct the PST Kernel
        self.pst_kernel = W * self.RHO * np.arctan(W * self.RHO) - 0.5 * np.log(
            1 + (W * self.RHO) ** 2
        )
        self.pst_kernel = S * self.pst_kernel / np.max(self.pst_kernel)

    def apply_kernel(self, sigma_LPF, thresh_min, thresh_max, morph_flag):
        """apply the phase kernel onto the image

        Args:
            sigma_LPF (float): std of the low pass filter
            thresh_min (float): minimum thershold, we keep features < thresh_min
            thresh_max (float): maximum thershold, we keep features > thresh_max
            morph_flag (boolean): whether apply morphological operation
        """
        self.img_denoised = denoise(img=self.img, rho=self.RHO, sigma_LPF=sigma_LPF)
        self.img_pst = ifft2(
            fft2(self.img_denoised) * fftshift(np.exp(-1j * self.pst_kernel))
        )
        self.pst_feature = np.angle(self.img_pst)
        if morph_flag == 0:
            self.pst_output = self.pst_feature
        else:
            self.pst_output = morph(
                img=self.img,
                feature=self.pst_feature,
                thresh_max=thresh_max,
                thresh_min=thresh_min,
            )

    def run(
        self,
        img_file,
        S,
        W,
        sigma_LPF,
        thresh_min,
        thresh_max,
        morph_flag,
    ):
        """wrap all steps of PST into a single run method

        Args:
            img_file (str): path to the image.
            S (float): phase strength of PST
            W (float): warp strength of PST
            sigma_LPF (float): std of the low pass filter
            thresh_min (float): minimum thershold, we keep features < thresh_min
            thresh_max (float): maximum thershold, we keep features > thresh_max
            morph_flag (boolean): whether apply morphological operation

        Returns:
            np.ndarray: PST output
        """
        self.load_img(img_file=img_file)
        self.init_kernel(S, W)
        self.apply_kernel(sigma_LPF, thresh_min, thresh_max, morph_flag)

        return self.pst_output



import cv2
import numpy as np
from numpy.fft import fft2, fftshift, ifft2

from .utils import cart2pol, denoise, morph


class PAGE:
    def __init__(self, direction_bins, h=None, w=None):
        """initialize the PAGE CPU version class

        Args:
            direction_bins (int): number of different diretions of edge to be extracted
            h (int, optional): height of the image to be processed. Defaults to None.
            w (int, optional): width of the image to be processed. Defaults to None.
        """
        self.h = h
        self.w = w
        self.direction_bins = direction_bins

    def load_img(self, img_file=None, img_array=None):
        """load the image from an ndarray or from an image file

        Args:
            img_file (str, optional): path to the image. Defaults to None.
            img_array (np.ndarray, optional): image in the form of np.ndarray. Defaults to None.
        """
        if img_array is not None:
            # directly load the image from numpy array
            self.img = img_array
        else:
            # load the image from the image file
            self.img = cv2.imread(img_file)
            # read the image size or resize to the indicated size (height x width)
            if not self.h and not self.w:
                self.h = self.img.shape[0]
                self.w = self.img.shape[1]
            else:
                self.img = cv2.imresize(self.img, [self.h, self.w])
            # convert to grayscale if it is RGB
        if self.img.ndim == 3:
            self.img = cv2.cvtColor(self.img, cv2.COLOR_BGR2GRAY)

    def init_kernel(self, mu_1, mu_2, sigma_1, sigma_2, S1, S2):
        """initialize the phase kernel of PST

        Args:
            mu_1 (float): Center frequency of a normal distributed passband filter ϕ1
            mu_2 (float):  Center frequency of log-normal  distributed passband filter ϕ2
            sigma_1 (float): Standard deviation of normal distributed passband filter ϕ1
            sigma_2 (float): Standard deviation of log-normal distributed passband filter ϕ2
            S1 (float): Phase strength of ϕ1
            S2 (float): Phase strength of ϕ2
        """
        # set the frequency grid
        u = np.linspace(-0.5, 0.5, self.h)
        v = np.linspace(-0.5, 0.5, self.w)
        [U, V] = np.meshgrid(u, v, indexing="ij")
        [self.THETA, self.RHO] = cart2pol(U, V)

        min_direction = np.pi / 180
        direction_span = np.pi / self.direction_bins
        directions = np.arange(min_direction, np.pi, direction_span)

        # create PAGE kernels channel by channel
        self.page_kernel = np.zeros([self.h, self.w, self.direction_bins])
        for i in range(self.direction_bins):
            tetav = directions[i]
            # Project onto new directionality basis for PAGE filter creation
            Uprime = U * np.cos(tetav) + V * np.sin(tetav)
            Vprime = -U * np.sin(tetav) + V * np.cos(tetav)

            # Create Normal component of PAGE filter
            Phi_1 = np.exp(-0.5 * ((abs(Uprime) - mu_1) / sigma_1) ** 2) / (
                1 * np.sqrt(2 * np.pi) * sigma_1
            )
            Phi_1 = (Phi_1 / np.max(Phi_1[:])) * S1

            # Create Log-Normal component of PAGE filter
            Phi_2 = np.exp(-0.5 * ((np.log(abs(Vprime)) - mu_2) / sigma_2) ** 2) / (
                abs(Vprime) * np.sqrt(2 * np.pi) * sigma_2
            )
            Phi_2 = (Phi_2 / np.max(Phi_2[:])) * S2

            # Add overall directional filter to PAGE filter array
            self.page_kernel[:, :, i] = Phi_1 * Phi_2

    def apply_kernel(self, sigma_LPF, thresh_min, thresh_max, morph_flag):
        """apply the phase kernel onto the image
        Args:
            sigma_LPF (float): std of the low pass filter
            thresh_min (float): minimum thershold, we keep features < thresh_min
            thresh_max (float): maximum thershold, we keep features > thresh_max
            morph_flag (boolean): whether apply morphological operation
        """

        # denoise on the loaded image
        self.img_denoised = denoise(img=self.img, rho=self.RHO, sigma_LPF=sigma_LPF)
        self.page_output = np.zeros([self.h, self.w, self.direction_bins])
        # apply the kernel channel by channel
        for i in range(self.direction_bins):
            self.img_page = ifft2(
                fft2(self.img_denoised)
                * fftshift(np.exp(-1j * self.page_kernel[:, :, i]))
            )
            self.page_feature = np.angle(self.img_page)
            # apply morphological operation if applicable
            if morph_flag == 0:
                self.page_output[:, :, i] = self.page_feature
            else:
                self.page_output[:, :, i] = morph(
                    img=self.img,
                    feature=self.page_feature,
                    thresh_max=thresh_max,
                    thresh_min=thresh_min,
                )

    def create_page_edge(self):
        """create results which color-coded directional edges"""
        # Create a weighted color image of PAGE output to visualize directionality of edges
        weight_step = 255 * 3 / self.direction_bins
        color_weight = np.arange(0, 255, weight_step)
        self.page_edge = np.zeros([self.h, self.w, 3])
        # step_edge = int(round(self.direction_bins/3))
        step_edge = self.direction_bins // 3
        for i in range(step_edge):
            self.page_edge[:, :, 0] = (
                color_weight[i] * self.page_output[:, :, i] + self.page_edge[:, :, 0]
            )
            self.page_edge[:, :, 1] = (
                color_weight[i] * self.page_output[:, :, i + step_edge]
                + self.page_edge[:, :, 1]
            )
            self.page_edge[:, :, 2] = (
                color_weight[i] * self.page_output[:, :, i + (2 * step_edge)]
                + self.page_edge[:, :, 2]
            )

        self.page_edge = (self.page_edge - np.min(self.page_edge)) / (
            np.max(self.page_edge) - np.min(self.page_edge)
        )

    def run(
        self,
        img_file,
        mu_1,
        mu_2,
        sigma_1,
        sigma_2,
        S1,
        S2,
        sigma_LPF,
        thresh_min,
        thresh_max,
        morph_flag,
    ):
        """wrap all steps of PAGE into a single run method

        Args:
            img_file (str): path to the image.
            mu_1 (float): Center frequency of a normal distributed passband filter ϕ1
            mu_2 (float):  Center frequency of log-normal  distributed passband filter ϕ2
            sigma_1 (float): Standard deviation of normal distributed passband filter ϕ1
            sigma_2 (float): Standard deviation of log-normal distributed passband filter ϕ2
            S1 (float): Phase strength of ϕ1
            S2 (float): Phase strength of ϕ2
            sigma_LPF (float): std of the low pass filter
            thresh_min (float): minimum thershold, we keep features < thresh_min
            thresh_max (float): maximum thershold, we keep features > thresh_max
            morph_flag (boolean): whether apply morphological operation

        Returns:
            np.ndarray: color-coded directional edge
        """
        # wrap load_img, init_kernel, apply_kernel, create_page_edge in one run
        self.load_img(img_file=img_file)
        self.init_kernel(mu_1, mu_2, sigma_1, sigma_2, S1, S2)
        self.apply_kernel(sigma_LPF, thresh_min, thresh_max, morph_flag)
        self.create_page_edge()

        return self.page_edge
